const 
  CARD8 = 1,
  CARD16 = 2,
  OFFSIZE = 1, //range = 1 - 4
  SID = 2; //range = 0 - 64999

var topDictOffset,
  topDictHash = {},
  privateDictHash = {},
  cffOffset;

//from ttx's cffLib.py
var topDictOperators = [
//opcode     name                  argument type   default    converter
	[0xc1e, 'ROS',        ['SID','SID','number'], null,      null],
	[0xc14, 'SyntheticBase',      'number',       null,      null],
	[0,        'version',            'SID',          null,      null],
	[1,        'Notice',             'SID',          null,      null],
	[0xc00,  'Copyright',          'SID',          null,      null],
	[2,        'FullName',           'SID',          null,      null],
	[0xc26, 'FontName',           'SID',          null,      null],
	[3,        'FamilyName',         'SID',          null,      null],
	[4,        'Weight',             'SID',          null,      null],
	[0xc01,  'isFixedPitch',       'number',       0,         null],
	[0xc02,  'ItalicAngle',        'number',       0,         null],
	[0xc03,  'UnderlinePosition',  'number',       null,      null],
	[0xc04,  'UnderlineThickness', 'number',       50,        null],
	[0xc05,  'PaintType',          'number',       0,         null],
	[0xc06,  'CharstringType',     'number',       2,         null],
	[0xc07,  'FontMatrix',         'array',  [0.001,0,0,0.001,0,0],  null],
	[13,       'UniqueID',           'number',       null,      null],
	[5,        'FontBBox',           'array',  [0,0,0,0],       null],
	[0xc08,  'StrokeWidth',        'number',       0,         null],
	[14,       'XUID',               'array',        null,      null],
	[0xc15, 'PostScript',         'SID',          null,      null],
	[0xc16, 'BaseFontName',       'SID',          null,      null],
	[0xc17, 'BaseFontBlend',      'delta',        null,      null],
	[0xc1f, 'CIDFontVersion',     'number',       0,         null],
	[0xc20, 'CIDFontRevision',    'number',       0,         null],
	[0xc21, 'CIDFontType',        'number',       0,         null],
	[0xc22, 'CIDCount',           'number',       8720,      null],
	[15,       'charset',            'number',       0,         null],
	[0xc23, 'UIDBase',            'number',       null,      null],
	[16,       'Encoding',           'number',       0,         null],
	[18,       'Private',       ['number','number'], null,      null],
	[0xc25, 'FDSelect',           'number',       null,      null],
	[0xc24, 'FDArray',            'number',       null,      null],
	[17,       'CharStrings',        'number',       null,      null],
];

var privateDictOperators = [
//opcode     name                  argument type   default    converter
	[6,        'BlueValues',         'delta',        null,      null],
	[7,        'OtherBlues',         'delta',        null,      null],
	[8,        'FamilyBlues',        'delta',        null,      null],
	[9,        'FamilyOtherBlues',   'delta',        null,      null],
	[0xc09,  'BlueScale',          'number',       0.039625,  null],
	[0xc0a, 'BlueShift',          'number',       7,         null],
	[0xc0b, 'BlueFuzz',           'number',       1,         null],
	[10,       'StdHW',              'number',       null,      null],
	[11,       'StdVW',              'number',       null,      null],
	[0xc0c, 'StemSnapH',          'delta',        null,      null],
	[0xc0d, 'StemSnapV',          'delta',        null,      null],
	[0xc0e, 'ForceBold',          'number',       0,         null],
	[0xc0f, 'ForceBoldThreshold', 'number',       null,      null],  // deprecated
	[0xc10, 'lenIV',              'number',       null,      null],  // deprecated
	[0xc11, 'LanguageGroup',      'number',       0,         null],
	[0xc12, 'ExpansionFactor',    'number',       0.06,      null],
	[0xc13, 'initialRandomSeed',  'number',       0,         null],
	[20,       'defaultWidthX',      'number',       0,         null],
	[21,       'nominalWidthX',      'number',       0,         null],
	[19,       'Subrs',              'number',       null,      null],
];

var allDictOperators = topDictOperators.concat(privateDictOperators);

function offsizeVal()
{
  return String.fromCharCode(rand_range(1, 5));
}

function sidVal()
{
  return intToString(rand_max(65000), SID);
}

function getDictValue(hash, key, descTable)
{
  if(key in hash)
    return hash[key];
  else
  {
    var desc;
    for(var i = 0; desc = descTable[i]; i++)
      if(desc[1] == key)
        return desc[3]; //default value
  }
  assert(1); //key not found
}

function parseCFFIndex(data, struct, containsDICT, hashToSaveTo /*only relevant if containsDICT = true*/)
{
  var start = struct.offset;
  var count = getMBInt(CARD16, data, struct.offset);
  struct.add(CARD16);
  var offsets = [];
  if(count) //if count is zero, the table is only 2 bytes long
  {
    var oSize = data.charCodeAt(struct.offset);
    struct.add(OFFSIZE, false, false, offsizeVal);
    
    for(var i = 0; i < count+1; i++) //starts with 0x01
    {
      offsets.push(getMBInt(oSize, data, struct.offset));
      struct.add(oSize);
    }
    for(var i = 1; i < count+1; i++)
    {
      if(containsDICT)
        parseCFFDict(data, struct, offsets[i] - offsets[i-1], hashToSaveTo);
      else
        struct.add(offsets[i] - offsets[i-1]);
    }
  }
}

function parseCFFCharset(data, struct, nGlyphs)
{
  struct.add(CARD8);
  var size;
  switch(data.charCodeAt(struct.offset-1))
  {
    case 0:
      struct.add(SID);
      size = SID;
      break;
    case 1:
      struct.add(SID);
      size = CARD8;
      break;
    case 2:
      struct.add(SID);
      size = CARD16;
      break;
  }
  for(var i = 0; i < nGlyphs-2; i++)
    struct.add(size);
}


function parseCFFDict(data, struct, length, hashToSaveTo)
{
  var endOffset = struct.offset + length;
  var operandStack = [];
  while(struct.offset < endOffset)
  {
    var value, b0 = data.charCodeAt(struct.offset);
    
    //operand(s)
    struct.add(1)
    if(b0 >= 32 && b0 <= 246)
    {
      value = b0 - 139;
      operandStack.push(value);
    }
    else if(b0 >= 247 && b0 <= 250)
    {
      value = (b0 - 247) * 256 + data.charCodeAt(struct.offset + 1) + 108;
      operandStack.push(value);
      struct.add(1);
    }
    else if(b0 >= 251 && b0 <= 254)
    {
      value = -(b0 - 251) * 256 - data.charCodeAt(struct.offset + 1) - 108;
      operandStack.push(value);
      struct.add(1);
    }
    else if(b0 == 28)
    {
      value = getMBInt(2, data, struct.offset);
      operandStack.push(value);
      struct.add(2);
    }
    else if(b0 == 29)
    {
      value = getMBInt(4, data, struct.offset);
      operandStack.push(value);
      struct.add(4);
    }
    else if(b0 == 30)
    {
      var keepGoing = true, i;
      for(i = 0; (data.charCodeAt(struct.offset + i) & 0xF) == 0xF; i++); //if second nibble is 0xF, the number ends
      //i dont think there's any need to set value or push the operand here
      struct.add(i); //todo: this should probably use bytes that don't contain a 0xf nibble until the end
    }
    else
      assert(data.charCodeAt(struct.offset) > 21); //unrecognized byte
    
    //operand added, check for operator
    if(data.charCodeAt(struct.offset) <= 21)
    {
      //operator
      var operator = data.charCodeAt(struct.offset);
      if(operator == 12)
      {
        operator = ls(operator, 8) + data.charCodeAt(struct.offset+1);
        struct.add(2);
      }
      else
        struct.add(1);
      
      var opArrItem, op, i;
      for(i = 0; opArrItem = allDictOperators[i]; i++)
        if(operator == opArrItem[0])
        {
          if(hashToSaveTo)
            hashToSaveTo[opArrItem[1]] = {
                                          operands: operandStack,
                                          desc: opArrItem
                                        };
          break;
        }
        
      assert(i == allDictOperators.length); //couldn't find the operator
      
      //clear the operand stack
      operandStack = [];
    }
  }
}

//encodeToHex(data.substring(offsets[i], offsets[i-1]))
function encodeToHex(str){
    var r="";
    var e=str.length;
    var c=0;
    var h;
    while(c<e){
        h=str.charCodeAt(c++).toString(16);
        while(h.length<3) h="0"+h;
        r+=h + " ";
    }
    return r;
}

function parseCFFHeader(data, struct)
{
  struct.add(CARD8, true); //major
  struct.add(CARD8, true); //minor
  struct.add(CARD8); //hdrSize
  struct.add(OFFSIZE, false, false, offsizeVal); //offSize
}

function parseCFFEncoding(data, struct)
{
  //todo: this code is currently untested
  
  var format = data.charCodeAt(struct.offset);
  
  assert(format > 1 && !(format & 0x80)); //probably means im parsing the font incorrectly
  
  struct.add(CARD8); //format
  
  switch(format & 0x7F)
  {
    case 0:
      var nCodes = data.charCodeAt(struct.offset);
      struct.add(CARD8); //nCodes
      for(var i = 0; i < nCodes; i++)
        struct.add(CARD8); //code
        
      break;
    case 1:
      var nRanges = data.charCodeAt(struct.offset);
      for(var i = 0; i < nRanges; i++)
      {
        struct.add(CARD8); //first
        struct.add(CARD8); //nLeft
      }
      break;
  }
  if(format & 0x80)
  {
    var count = data.charCodeAt(struct.offset);
    for(var i = 0; i < count; i++)
    {
      struct.add(CARD8);
      struct.add(SID);
    }
  }
}

function handlePrivateDictEntries(data, struct)
{
  //no need to save the offset here, since it will happen at the end of parsing the top dict
  var value;
  if("Subrs" in privateDictHash)
  {
    value = privateDictHash["Subrs"].operands.pop();
    if(value)
    {
      //struct.offset should currently be private dict offset 
      struct.offset += value;
      parseCFFIndex(data, struct);
    }
  }
}

function handleTopDictEntries(data, struct)
{
  var value,
    currOffset = struct.offset,
    nGlyphs;
    
  if("CharStrings" in topDictHash)
  {
    value = topDictHash["CharStrings"].operands.pop();
    if(value)
    {
      struct.offset = cffOffset + value; //offset
      nGlyphs = getMBInt(CARD16, data, struct.offset);
      parseCFFIndex(data, struct);
      if("charset" in topDictHash)
      {
        value = topDictHash["charset"].operands.pop(); 
        if(value)
        {
          struct.offset = cffOffset + value; //offset
          parseCFFCharset(data, struct, nGlyphs);
        }
      }
    }
  }
  if("Encoding" in topDictHash)
  {
    value = topDictHash["Encoding"].operands.pop();
    if(value)
    {
      struct.offset = cffOffset + value; //offset
      parseCFFEncoding(data, struct);
    }
  }
  if("Private" in topDictHash)
  {
    value = topDictHash["Private"].operands.pop();
    if(value)
    {
      var length = topDictHash["Private"].operands.pop();
      struct.offset = cffOffset + value;
      parseCFFDict(data, struct, length, privateDictHash);
      //local subrs relies on an offset relative to the start of the private dict
      struct.offset = cffOffset + value;
      handlePrivateDictEntries(data, struct);
    }
  }
  if("FDArray" in topDictHash)
  {
    //todo: implement this and FDSelect
  }
  struct.offset = currOffset;
  
}

function parseCFF(data, struct)
{
  cffOffset = struct.offset;
  parseCFFHeader(data, struct);
  parseCFFIndex(data, struct); //name
  parseCFFIndex(data, struct, true, topDictHash); //top dict
  handleTopDictEntries(data, struct);
  
  parseCFFIndex(data, struct); //string
  parseCFFIndex(data, struct); //global subr
}